import pygame
import random
import math


class Hitbox:
    def __init__(self, x: float, y: float, width: float, height: float) -> None:
        self.x: float = x
        self.y: float = y
        self.width: float = width
        self.height: float = height

    def collides(self, other: "Hitbox") -> bool:
        return (
            self.x < other.x + other.width
            and self.x + self.width > other.x
            and self.y < other.y + other.height
            and self.y + self.height > other.y
        )


class Crow:
    def __init__(self) -> None:
        self.x: float = 400
        self.y: float = 300
        self.x_speed: float = 200
        self.y_speed: float = 0
        self.x_speed_max: float = 400
        self.y_speed_max: float = 200
        self.x_accel: float = 0
        self.y_accel: float = 0
        self.x_accel_max: float = 2000
        self.y_accel_max: float = 2000
        self.x_accel_change: float = 2000
        self.y_accel_change: float = 2000
        self.health = 3
        self.scale = 10
        self.image = pygame.image.load("assets/crow.png")
        self.image = pygame.transform.scale(
            self.image,
            (
                self.image.get_width() * self.scale,
                self.image.get_height() * self.scale
            )
        )
        self.width = self.image.get_width()
        self.height = self.image.get_height()

    def update(self, dt: float) -> None:
        self.x += self.x_speed * dt
        self.y += self.y_speed * dt

        self.x_speed += self.x_accel * dt
        self.y_speed += self.y_accel * dt

        self.x_accel += self.x_accel_change * dt * random.uniform(-1, 1)
        self.y_accel += self.y_accel_change * dt * random.uniform(-1, 1)

        if self.x_accel > self.x_accel_max:
            self.x_accel = self.x_accel_max
        elif self.x_accel < -self.x_accel_max:
            self.x_accel = -self.x_accel_max

        if self.x_speed > self.x_speed_max:
            self.x_speed = self.x_speed_max
            self.x_accel = 0
        elif self.x_speed < -self.x_speed_max:
            self.x_speed = -self.x_speed_max
            self.x_accel = 0

        if self.y_speed > self.y_speed_max:
            self.y_speed = self.y_speed_max
            self.y_accel = 0
        elif self.y_speed < -self.y_speed_max:
            self.y_speed = -self.y_speed_max
            self.y_accel = 0

        if self.x < 0:
            self.x = 0
            self.x_speed *= -1
        elif self.x > 800 - self.width:
            self.x = 800 - self.width
            self.x_speed *= -1

        if self.y < 0:
            self.y = 0
            self.y_speed *= -1
        elif self.y > 600 - self.height - 200:
            self.y = 600 - self.height - 200
            self.y_speed *= -1

    def hitbox(self) -> Hitbox:
        return Hitbox(self.x, self.y, self.width, self.height)


class Bullet:
    def __init__(self, x: float, y: float) -> None:
        self.x: float = x
        self.y: float = y
        self.y_speed: float = -500
        self.scale = 10
        self.image = pygame.image.load("assets/bullet.png")
        self.image = pygame.transform.scale(
            self.image,
            (
                self.image.get_width() * self.scale,
                self.image.get_height() * self.scale
            )
        )
        self.width = self.image.get_width()
        self.height = self.image.get_height()

    def update(self, dt: float) -> None:
        self.y += self.y_speed * dt

    def hitbox(self) -> Hitbox:
        return Hitbox(self.x, self.y, self.width, self.height)


class Player:
    def __init__(self) -> None:
        self.x: float = 400
        self.y: float = 500
        self.x_speed: float = 0
        self.x_speed_max: float = 300
        self.x_accel: float = 1200
        self.scale = 10
        self.image = pygame.image.load("assets/turret.png")
        self.image = pygame.transform.scale(
            self.image,
            (
                self.image.get_width() * self.scale,
                self.image.get_height() * self.scale
            )
        )
        self.width = self.image.get_width()
        self.height = self.image.get_height()
        self.shot_cooldown = 0.5
        self.shot_timer: float = 0
        self.bullets = 6

    def update(self, dt: float) -> None:
        self.x += self.x_speed * dt

        if self.x_speed > self.x_speed_max:
            self.x_speed = self.x_speed_max
        elif self.x_speed < -self.x_speed_max:
            self.x_speed = -self.x_speed_max

        if self.x < 0:
            self.x = 0
            self.x_speed *= -1
        elif self.x > 800 - self.width:
            self.x = 800 - self.width
            self.x_speed *= -1

    def press_left(self, dt: float) -> None:
        self.x_speed -= self.x_accel * dt

    def press_right(self, dt: float) -> None:
        self.x_speed += self.x_accel * dt

    def stop(self, dt: float) -> None:
        if abs(self.x_speed) < self.x_accel * dt:
            self.x_speed = 0
        elif self.x_speed != 0:
            self.x_speed += -self.x_speed / abs(self.x_speed) * self.x_accel * dt

    def shoot(self, time: float) -> Bullet | None:
        if time - self.shot_timer < self.shot_cooldown or self.bullets == 0:
            return None

        self.shot_timer = time
        self.bullets -= 1
        return Bullet(self.x + self.width / 2, self.y)


class Game:
    def __init__(self) -> None:
        pygame.init()
        self.screen = pygame.display.set_mode((800, 600))
        self.running = True
        self.clock = pygame.time.Clock()

        self.player = Player()
        self.crow = Crow()
        self.bullets: list[Bullet] = []

        self.time_start = pygame.time.get_ticks()
        self.time_limit = 60

        # monospace
        self.font = pygame.font.SysFont("monospace", 20)
        self.end_font = pygame.font.Font(None, 100)
        self.end_message: str | None = None

    def update(self, dt: float) -> None:
        self.player.update(dt)
        self.crow.update(dt)

        for bullet in self.bullets:
            bullet.update(dt)
            if bullet.hitbox().collides(self.crow.hitbox()):
                self.crow.health -= 1
                self.bullets.remove(bullet)
            if bullet.y < -bullet.height:
                self.bullets.remove(bullet)

        if self.crow.health <= 0:
            self.end_message = "You Win!"
            self.running = False

        if pygame.time.get_ticks() - self.time_start > self.time_limit * 1000 \
                or (self.player.bullets == 0 and len(self.bullets) == 0):
            self.end_message = "You Lose!"
            self.running = False

    def draw(self, time: float) -> None:
        self.screen.fill((255, 255, 255))
        self.screen.blit(self.player.image, (self.player.x, self.player.y))
        self.screen.blit(self.crow.image, (self.crow.x, self.crow.y))

        for bullet in self.bullets:
            self.screen.blit(bullet.image, (bullet.x, bullet.y))

        text = self.font.render(f"Health: {self.crow.health}", True, (0, 0, 0))
        self.screen.blit(text, (800 - text.get_width() - 10, 10))
        text = self.font.render(f"Bullets: {self.player.bullets}", True, (0, 0, 0))
        self.screen.blit(text, (800 - text.get_width() - 10, 40))
        text = self.font.render(f"Time: {math.floor(self.time_limit - time)}", True, (0, 0, 0))
        self.screen.blit(text, (800 - text.get_width() - 10, 70))

        texts = [
            self.font.render("Controls:", True, (0, 0, 0)),
            self.font.render("A/Left: move left", True, (0, 0, 0)),
            self.font.render("D/Right: move right", True, (0, 0, 0)),
            self.font.render("Space/Up: shoot", True, (0, 0, 0))
        ]
        for i, text in enumerate(texts):
            self.screen.blit(text, (10, 10 + i * 30))

        text = self.font.render(
            f"Time: {math.floor(self.time_limit - time)}",
            True,
            (255, 255, 255)
        )

        if self.end_message:
            text = self.end_font.render(self.end_message, True, (255, 0, 0))
            self.screen.blit(text, (400 - text.get_width() / 2, 300 - text.get_height() / 2))

        pygame.display.flip()


def main() -> None:
    game = Game()

    while game.running:
        dt = game.clock.tick(60) / 1000
        time = pygame.time.get_ticks() / 1000

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                return

        keys = pygame.key.get_pressed()
        if keys[pygame.K_LEFT] or keys[pygame.K_a]:
            game.player.press_left(dt)
        elif keys[pygame.K_RIGHT] or keys[pygame.K_d]:
            game.player.press_right(dt)
        else:
            game.player.stop(dt)
        if keys[pygame.K_SPACE] or keys[pygame.K_UP]:
            bullet = game.player.shoot(time)
            if bullet is not None:
                game.bullets.append(bullet)

        game.update(dt)
        game.draw(time)

    while True:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                return


if __name__ == "__main__":
    main()
